@@ -47,13 +47,16 @@ typedef unsigned int uint;
 # ifndef  asm
 #  define asm			__asm__
 # endif
+#endif
 # ifndef  extension
 #  define extension		__extension__
 # endif
-#endif
 #ifndef  attribute
 # define attribute(attr)	__attribute__(attr)
 #endif
+#ifndef offsetof
+# define offsetof(type,memb)	__builtin_offsetof(type,memb)
+#endif
 
 /*
  * This is lent from the kernel by e.g. using
@@ -64,8 +67,7 @@ typedef unsigned int uint;
  *
  * on the appropiate architecture (here on i686 for i586).
  */
-static inline void prefetch(const void *restrict x) attribute((used,always_inline));
-static inline void prefetch(const void *restrict x)
+extern inline void attribute((used,__gnu_inline__,always_inline,)) prefetch(const void *restrict x)
 {
 #if   defined(__x86_64__)
     asm volatile ("prefetcht0 %0"  :: "m" (*(unsigned long *)x))
@@ -91,6 +93,8 @@ static inline void prefetch(const void *restrict x)
 		  "   \n664:\n"
 		  ".previous"
 		  :: "i" ((0*32+25)), "r" (x))
+#else
+    __builtin_prefetch ((x), 0, 1);
 #endif
     ;
 }
@@ -158,6 +162,18 @@ static inline void delete(list_t *restrict entry)
     initial(entry);
 }
 
+/*
+ * Replace an entry by a new one.
+ */
+static inline void replace(list_t *restrict old, list_t *restrict new) attribute((always_inline,nonnull(1,2)));
+static inline void replace(list_t *restrict old, list_t *restrict new)
+{
+    new->next = old->next;
+    new->next->prev = new;
+    new->prev = old->prev;
+    new->prev->next = new;
+}
+
 static inline void join(list_t *restrict list, list_t *restrict head) attribute((always_inline,nonnull(1,2)));
 static inline void join(list_t *restrict list, list_t *restrict head)
 {
@@ -165,22 +181,40 @@ static inline void join(list_t *restrict list, list_t *restrict head)
 
     if (first != list) {
 	list_t * last = list->prev;
-       	list_t * at = head->next;
+	list_t * at = head->next;
 
-       	first->prev = head;
-       	head->next = first;
+	first->prev = head;
+	head->next = first;
 
-       	last->next = at;
-       	at->prev = last;
+	last->next = at;
+	at->prev = last;
     }
 }
 
-static inline boolean list_empty(list_t *restrict head) attribute((always_inline,nonnull(1)));
-static inline boolean list_empty(list_t *restrict head)
+static inline boolean list_empty(const list_t *restrict const head) attribute((always_inline,nonnull(1)));
+static inline boolean list_empty(const list_t *restrict const head)
 {
      return head->next == head;
 }
 
+static inline void move_head(list_t *restrict entry, list_t *restrict head) attribute((always_inline,nonnull(1,2)));
+static inline void move_head(list_t *restrict entry, list_t *restrict head)
+{
+    list_t * prev = entry->prev;
+    list_t * next = entry->next;
+
+    next->prev = prev;		/* remove enty from old list */
+    prev->next = next;
+
+    prev = head;
+    next = head->next;
+
+    next->prev = entry;		/* and add it at head of new list */
+    entry->next = next;
+    entry->prev = prev;
+    prev->next = entry;
+}
+
 static inline void move_tail(list_t *restrict entry, list_t *restrict head) attribute((always_inline,nonnull(1,2)));
 static inline void move_tail(list_t *restrict entry, list_t *restrict head)
 {
@@ -294,12 +328,12 @@ extern const char *const delimeter;
 extern void error(const char *restrict fmt, ...) attribute((noreturn,format(printf,1,2)));
 extern void warn (const char *restrict fmt, ...) attribute((format(printf,1,2)));
 extern void info (int level, const char *restrict fmt, ...) attribute((format(printf,2,3)));
-extern inline int map_has_runlevels(void) attribute((always_inline));
-extern inline char map_runlevel_to_key(const int runlevel);
-extern inline ushort map_key_to_lvl(const char key);
-extern inline const char *map_runlevel_to_location(const int runlevel);
-extern inline ushort map_runlevel_to_lvl(const int runlevel);
-extern inline ushort map_runlevel_to_seek(const int runlevel);
+int map_has_runlevels(void);
+char map_runlevel_to_key(const int runlevel);
+ushort map_key_to_lvl(const char key);
+const char *map_runlevel_to_location(const int runlevel);
+ushort map_runlevel_to_lvl(const int runlevel);
+ushort map_runlevel_to_seek(const int runlevel);
 extern ushort str2lvl(const char *restrict lvl) attribute((nonnull(1)));
 extern char * lvl2str(const ushort lvl);
 
@@ -314,29 +348,30 @@ static inline char * xstrdup(const char *restrict s)
     return r;
 } 
 
+extern char empty[];
 #define xreset(ptr)	\
-	{char *restrict tmp = (char *restrict)ptr; if (ptr && *tmp) free(ptr);} ptr = NULL
+	{ if (ptr && empty != ptr) free(ptr);} ptr = NULL
 
 #if defined(HAS_unlinkat) && defined(_ATFILE_SOURCE) && !defined(__stub_unlinkat)
 # define xremove(d,x) (__extension__ ({ if ((dryrun ? 0 : \
 	(unlinkat(d,x,0) != 0 && (errno != EISDIR || unlinkat(d,x,AT_REMOVEDIR) != 0)))) \
-	warn ("can not remove(%s%s): %s\n", rcd, x, strerror(errno)); \
+	warn ("can not remove(%s/%s%s): %s\n", path, rcd, x, strerror(errno)); \
 	else \
 	info(1, "remove service %s/%s%s\n", path, rcd, x); }))
 #else
 # define xremove(d,x) (__extension__ ({ if ((dryrun ? 0 : (remove(x) != 0))) \
-	warn ("can not remove(%s%s): %s\n", rcd, x, strerror(errno)); \
+	warn ("can not remove(%s/%s%s): %s\n", path, rcd, x, strerror(errno)); \
 	else \
 	info(1, "remove service %s/%s%s\n", path, rcd, x); }))
 #endif
 #if defined(HAS_symlinkat) && defined(_ATFILE_SOURCE) && !defined(__stub_symlinkat)
 # define xsymlink(d,x,y) (__extension__ ({ if ((dryrun ? 0 : (symlinkat(x, d, y) != 0))) \
-	warn ("can not symlink(%s, %s%s): %s\n", x, rcd, y, strerror(errno)); \
+	warn ("can not symlink(%s, %s/%s%s): %s\n", x, path, rcd, y, strerror(errno)); \
 	else \
 	info(1, "enable service %s -> %s/%s%s\n", x, path, rcd, y); }))
 #else
 # define xsymlink(d,x,y) (__extension__ ({ if ((dryrun ? 0 : (symlink(x, y) != 0))) \
-	warn ("can not symlink(%s, %s%s): %s\n", x, rcd, y, strerror(errno)); \
+	warn ("can not symlink(%s, %s/%s%s): %s\n", x, path, rcd, y, strerror(errno)); \
 	else \
 	info(1, "enable service %s -> %s/%s%s\n", x, path, rcd, y); }))
 #endif
@@ -379,6 +414,9 @@ static inline char * xstrdup(const char *restrict s)
 #define SERV_NOSTOP	0x0100
 #define SERV_CMDLINE	0x0200
 #define SERV_FIRST	0x0400
+#define SERV_ENFORCE	0x0800
+#define SERV_WARNED	0x1000
+#define SERV_SYSTEMD	0x2000
 
 /*
  * Bits of the runlevels
@@ -416,3 +454,37 @@ static inline char * xstrdup(const char *restrict s)
  * Maximum start/stop level
  */
 #define MAX_DEEP 99
+
+#ifdef NEED_RUNLEVELS_DEF
+static struct {
+    char *location;
+    const ushort lvl;
+    const ushort seek;
+    const char key;
+} attribute((aligned(sizeof(char*)))) runlevel_locations[] = {
+#ifdef SUSE     /* SuSE's SystemV link scheme */
+    {"rc0.d/",    LVL_HALT,   LVL_NORM, '0'},
+    {"rc1.d/",    LVL_ONE,    LVL_NORM, '1'}, /* runlevel 1 switch over to single user mode */
+    {"rc2.d/",    LVL_TWO,    LVL_NORM, '2'},
+    {"rc3.d/",    LVL_THREE,  LVL_NORM, '3'},
+    {"rc4.d/",    LVL_FOUR,   LVL_NORM, '4'},
+    {"rc5.d/",    LVL_FIVE,   LVL_NORM, '5'},
+    {"rc6.d/",    LVL_REBOOT, LVL_NORM, '6'},
+    {"rcS.d/",    LVL_SINGLE, LVL_NORM, 'S'}, /* runlevel S is for single user mode */
+    {"boot.d/",   LVL_BOOT,   LVL_BOOT, 'B'}, /* runlevel B is for system initialization */
+#else           /* not SUSE (actually, Debian) */
+    {"../rc0.d/", LVL_HALT,   LVL_NORM, '0'},
+    {"../rc1.d/", LVL_ONE,    LVL_NORM, '1'}, /* runlevel 1 switch over to single user mode */
+    {"../rc2.d/", LVL_TWO,    LVL_NORM, '2'},
+    {"../rc3.d/", LVL_THREE,  LVL_NORM, '3'},
+    {"../rc4.d/", LVL_FOUR,   LVL_NORM, '4'},
+    {"../rc5.d/", LVL_FIVE,   LVL_NORM, '5'},
+    {"../rc6.d/", LVL_REBOOT, LVL_NORM, '6'},
+    {"../rcS.d/", LVL_BOOT,   LVL_BOOT, 'S'}, /* runlevel S is for system initialization */
+                /* On e.g. Debian there exist no boot.d */
+#endif          /* not SUSE */
+};
+
+
+#define RUNLEVELS (int)(sizeof(runlevel_locations)/sizeof(runlevel_locations[0]))
+#endif     /* done checking if we need runlevels defined */
